//----------------------------------------------------------------------------
// Code taken from `https://github.com/blastur/magboot` and partialy modified.
//----------------------------------------------------------------------------
#include <inttypes.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/boot.h>
#include <stdbool.h>
#include <stdint.h>
#include <util/delay.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <avr/sfr_defs.h>

#include "main.h"
#include "uart.h"

//----------------------------------------------------------------------------
void ports_init(void);
uint16_t checksum(uint8_t *data, size_t size);
bool cmd_write_page(uint16_t *addr);
bool cmd_load_addr(uint16_t *addr);
bool cmd_device_id(void);
bool cmd_reset(void);
void jump(uint16_t addr);

//----------------------------------------------------------------------------
int main(void)
{
  uint16_t addr = 0;
  bool fail = false;

  // initialize ports
  ports_init();
  // initialize UART
  uart_init();

  if (bit_is_clear(MCUSR, EXTRF) || bit_is_set(MCUSR, PORF)) {
    // Bypass Bootloader if reset caused by watchdog, power-on or brown-out
    MCUSR &= ~(_BV(PORF));
    jump(0);
  }
  else
    MCUSR &= ~(_BV(EXTRF));

  wdt_enable(WDTO_8S);

  // blink with "Bank select" and "Card select" LEDs
  switch_memory_bank_0();
  unselect_card();
  _delay_ms(250);
  switch_memory_bank_1();
  select_card();
  _delay_ms(250);
  switch_memory_bank_0();
  unselect_card();
  _delay_ms(250);
  switch_memory_bank_1();
  select_card();
  _delay_ms(250);
  unselect_card();
  _delay_ms(1500);

  while (1) {
    switch (uart_read_byte()) {
      // Device ID
      case 'I':
        fail = cmd_device_id();
        break;

      // Address
      case 'A':
        fail = cmd_load_addr(&addr);
        break;

      // Write page
      case 'W':
        fail = cmd_write_page(&addr);
        break;

      // Reset
      case 'R':
        fail = cmd_reset();
        break;

      default:
        fail = true;
        break;
    }

    if (fail)
      CMD_FAIL();
    else
      CMD_OK();

    wdt_reset();
  }
}
//----------------------------------------------------------------------------
void ports_init(void)
{
  MEM_A14_DIR |= MEM_A14_BIT;     // output
  MEM_A14_PORT &= ~MEM_A14_BIT;   // A14=0

  MEM_ALE_DIR |= MEM_ALE_BIT;     // output
  MEM_ALE_PORT &= ~MEM_ALE_BIT;   // ALE=0

  MEM_CSW_DIR |= MEM_CSW_BIT;     // output
  MEM_CSW_PORT |= MEM_CSW_BIT;    // -CSW=1

  MEM_WEM_DIR |= MEM_WEM_BIT;     // output
  MEM_WEM_PORT |= MEM_WEM_BIT;    // -WEM=1

  // Data Bus to High Impedance State
  DATABUS_L_DIR &= ~DATABUS_L_MASK;   // Data Bus as Input
  DATABUS_H_DIR &= ~DATABUS_H_MASK;
  DATABUS_L_DATA &= ~DATABUS_L_MASK;  // Z-State
  DATABUS_H_DATA &= ~DATABUS_H_MASK;

  // configure_pin_available
  DDRB &= ~(1 << DDB0);
  PORTB |= (1 << PORTB0);
  // configure_pin_locked
  DDRB &= ~(1 << DDB1);
  PORTB |= (1 << PORTB1);

  // enable outputs for MOSI, SCK, SS, input for MISO
  DDRB |= (1 << DDB5);  // SCK
  DDRB &= ~(1 << DDB4); // MISO
  DDRB |= (1 << DDB3);  // MOSI
  DDRB |= (1 << DDB2);  // SS

  // unselect_card
  PORTB |= (1 << PORTB2);
}
//----------------------------------------------------------------------------
uint16_t checksum(uint8_t *data, size_t size)
{
  uint16_t *words = (uint16_t *) data;
  size_t num_words = size / 2;
  uint32_t sum = 0;
  uint8_t i;

  for (i = 0; i < num_words; i++)
    sum += words[i];

  // Fold
  while (sum >> 16)
    sum = (sum & 0xFFFF) + (sum >> 16);

  return (uint16_t) sum;
}
//----------------------------------------------------------------------------
bool cmd_write_page(uint16_t *addr)
{
  uint16_t i;
  uint16_t page = *addr;
  uint8_t buf[SPM_PAGESIZE];
  uint16_t expected_csum, actual_csum;

  boot_page_erase(page);

  expected_csum = uart_read_byte();
  expected_csum += uart_read_byte() << 8;

  for(i = 0; i < SPM_PAGESIZE; i++)
    buf[i] = uart_read_byte();

  actual_csum = checksum(buf, sizeof(buf));

  if (expected_csum != actual_csum)
    return true;

  boot_spm_busy_wait();
  for (i = 0; i < SPM_PAGESIZE; i += 2) {
    uint16_t w = buf[i];
    w += buf[i + 1] << 8;
    boot_page_fill(page + i, w);
  }

  boot_page_write(page);
  boot_spm_busy_wait();

  boot_rww_enable();

  // Auto-increment address
  *addr += SPM_PAGESIZE;

  return false;
}
//----------------------------------------------------------------------------
bool cmd_load_addr(uint16_t *addr)
{
  // 16-bit addr, little-endian
  *addr = uart_read_byte();
  *addr += uart_read_byte() << 8;

  return false;
}
//-----------------------------------------------------------------------------
bool cmd_device_id(void)
{
  bool fail = false;

  // Always read full signature, even if fail is detected
  if (uart_read_byte() != SIGNATURE_0)
    fail = true;
  if (uart_read_byte() != SIGNATURE_1)
    fail = true;
  if (uart_read_byte() != SIGNATURE_2)
    fail = true;

  return fail;
}
//-----------------------------------------------------------------------------
bool cmd_reset(void)
{
  wdt_enable(WDTO_15MS);
  while (1)
    ;

  return false; // Unreachable
}
//----------------------------------------------------------------------------
void jump(uint16_t addr)
{
  jump_t func;

  /* Clear WDRF to prevent "infinite reset loop". Do not rely on application
   * to clear it, since it may not be WDT-aware and cannot account for the WDT
   * usage of Bootloader. The downside of this is however that an application can
   * never detect WDT reset during initialization.
   */
  MCUSR &= ~(_BV(WDRF));
  wdt_disable();
  func = (jump_t) addr;
  func();
}
//-----------------------------------------------------------------------------
