//----------------------------------------------------------------------------
// Project: PP 01 SD ROM Module
// Author:  RomBor, rombor18 <at> gmail.com
// Date:    02.12. 2018
// MPU:     ATmega328P@16Mhz
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
#include <inttypes.h>
#include <string.h>
#include <ctype.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <avr/eeprom.h>

#include "main.h"
#include "sd_raw.h"
#include "partition.h"
#include "fat.h"
#include "timer.h"

//----------------------------------------------------------------------------
// global variables
static uint8_t buffer[BUFFER_SIZE];     // general and sector buffer
static struct fat_fs_struct *pFS;       // file system
static struct partition_struct *pPart;  // disk partition
struct fat_file_struct *pWorkRdFile;    // working file opened for reading
struct fat_file_struct *pWorkWrFile;    // working file opened for writing
static uint8_t crc, val, val2, *pb;
static char cur_dir[FAT_MAX_PATH], file_mask[FAT_MAX_FILE_NAME_LEN], *pFile_mask_ext, *p;


// strings in Program memory
const char pstr_mngr_file_name[] PROGMEM = "ppsdrom.rmm";

//----------------------------------------------------------------------------
// function prototypes
uint8_t set_file_attrs(const char *pFilePath, struct fat_dir_entry_struct *pDE, uint8_t new_attrs);
uint8_t open_file_for_write(const char *pFilePath);
uint8_t open_file_for_read(const char *pFilePath);
uint8_t prepare_rom_file_into_memory(const char *pRomFilePath, uint16_t addr);
uint8_t prepare_manager_into_memory();
uint8_t store_file_into_memory(struct fat_dir_entry_struct *dir_entry, uint16_t addr, uint16_t count, uint16_t *pWritten);
uint8_t memory_write_file_content(struct fat_file_struct *pFD, uint16_t addr, uint16_t count, uint16_t *pWritten);
void memory_init(void);
void disable_memory_write(void);
void enable_memory_write(void);
void memory_write_byte(uint16_t addr, uint8_t bt);

void prepare_sd_card(void);
void close_sd_card(void);

void prepare_file_name_mask();
uint8_t contains_special_chars(const char *pName, uint8_t wildcards, uint8_t slash, uint8_t unallowed);
uint8_t match_name(const char *pMask, const char *pName);

void uart_init(void);
uint8_t read_byte(uint8_t *bt, uint32_t to_ms);
uint8_t read_byte_ex(uint8_t *bt);
uint8_t read_byte_crc_check(void);
void send_byte(uint8_t bt);

void config_read_current_path(char *path);
void config_write_current_path(char *path);

// external functions
#define memory_write_addr_byte sd_raw_send_byte
extern void handleCommand(void);

//----------------------------------------------------------------------------
int main(void)
{
  uint8_t command, length = 0, last_block;
  int8_t op, err, manager_loaded;
  uint16_t addr = 0, val16;
  int16_t vals16;
  struct fat_dir_struct *pDS, *pDSnew;
  struct fat_dir_entry_struct de, *pDE;

  // we will just use ordinary idle mode
  set_sleep_mode(SLEEP_MODE_IDLE);

  // initialize Timer
  timer_init();
  // initialize UART
  uart_init();
  // initialize memory ctrl pins
  memory_init();
  // configure SD card port pins
  sd_raw_init();

  sei();  // enable interrupt

  manager_loaded = 0;
  pPart = NULL;
  pFS = NULL;
  pDS = NULL;
  pDSnew = NULL;
  pWorkRdFile = NULL;
  pWorkWrFile = NULL;
  *cur_dir = '\0';
  *file_mask = '\0';
  pFile_mask_ext = NULL;

  while (1) {
    // wait for inserting of card
    prepare_sd_card();

    if (!manager_loaded)
      manager_loaded = (prepare_manager_into_memory() == RESULT_OK);

    op = OP_NOTHING;
    err = RESULT_OK;

    while (1) {
      if (!sd_raw_available())  // card removed "on the fly" !!!
        break;

      if (!manager_loaded) {
        // blink with "Bank select" LED when Manager is not loaded successfully
        switch_memory_bank_0();
        _delay_ms(250);
        switch_memory_bank_1();
        _delay_ms(250);
        switch_memory_bank_0();
        _delay_ms(250);
        switch_memory_bank_1();
        _delay_ms(1000);
        continue;
      }

      if (op == OP_NOTHING) {
        // read last used path on SD card
        config_read_current_path(cur_dir);

        val = 0;
        if (*cur_dir != '\0') {
          val = fat_get_dir_entry_of_path(pFS, cur_dir, &de);
          if (!val) {
            *cur_dir = '\0';
            config_write_current_path(cur_dir);
          }
        }
        if (!val || *cur_dir == '\0')
          fat_get_dir_entry_of_path(pFS, "/", &de);
        pDS = fat_open_dir(pFS, &de);

        op = OP_NO_TRANSFER;
      }

      if (!read_byte(&crc, WAIT_SYNCHRO_BYTE)) // wait synchro byte
        continue;
      if (crc != SYNCHRO_BYTE)
        continue;

      if (!read_byte(&command, WAIT_COMMAND_TIMEOUT)) // wait command
        continue;
      crc ^= command;

      op = OP_NO_TRANSFER;
      err = RESULT_OK;
      switch (command) {
        case '@' :  // FW version
        case '0' :  // set SRAM Bank 0
        case '1' :  // set SRAM Bank 1
        case 'K' :  // get current SD card path
        case 'P' :  // prepare next 4 kB from opened file
          if (read_byte_crc_check()) {
            if (command != '0' && command != '1')
              op = OP_SEND_DATA; // send data
          }
          else
            op = OP_ERROR;
          break;

        case 'L' :  // catalogue list
        case 'O' :  // open file for reading
        case 'o' :  // open file for reading
        case 'A' :  // get/set file attributes
        case 'a' :  // get/set file attributes
          op = OP_RECEIVE_DATA | OP_SEND_DATA; // receive and send data
          break;

        case 'M' :  // change directory
        case 'D' :  // delete file(s) in current directory
        case 'd' :  // delete empty directory
        case 'F' :  // create new directory
        case 'N' :  // rename file/directory
        case 'R' :  // insert file into ROM Module
        case 'r' :  // insert file into ROM Module
        case 'C' :  // create and open file for saving
        case 'c' :  // create and open file for saving
        case 'B' :  // write block to file
          op = OP_RECEIVE_DATA; // receive data
          break;

        default :
          op = OP_ERROR;
          break;
      }

      if (op == OP_ERROR) {   // ** ERROR **
        send_byte(NAK);   // send NAK
      }
      else if (op == OP_NO_TRANSFER) {     // ** WITHOUT ANY DATA TRANSFER **
        if (command == '0')
          switch_memory_bank_0();
        else if (command == '1')
          switch_memory_bank_1();

        send_byte(err);   // send ERR code
      }
      else {
        if (op & OP_RECEIVE_DATA) {    // ** RECEIVE DATA **
          if (command == 'B') { // write block to file
            if (!read_byte_ex(&last_block))
              op = OP_ERROR;
          }

          if (op != OP_ERROR && read_byte_ex(&val)) { // read string/block length
            length = val;
            val16 = (uint16_t) val;
            if (command == 'L' && val >= FAT_MAX_FILE_NAME_LEN) // catalogue list
              val16 = 0;
            else if (command == 'B' && val == 0) // write block to file
              val16 = 256;
            pb = buffer;
            while (val16) {
              if (!read_byte_ex(&val2)) {
                op = OP_ERROR;
                break;
              }
              *pb++ = val2;
              val16--;
            }
            *pb = '\0';

            if (op != OP_ERROR) {
              if (command == 'N') { // rename file/directory
                if (read_byte_ex(&val)) { // new file/directory name
                  pb = buffer + 256;
                  while (val) {
                    if (!read_byte_ex(&val2)) {
                      op = OP_ERROR;
                      break;
                    }
                    *pb++ = val2;
                    val--;
                  }
                  *pb = '\0';
                  val = pb - (buffer + 256); // new name length
                }
                else
                  op = OP_ERROR;
              }
              else if (command == 'R' || command == 'r') { // insert file into ROM Module
                if (read_byte_ex(&val) && read_byte_ex(&val2))
                  addr = (val2 << 8) | val;
                else
                  op = OP_ERROR;
              }
              else if (command == 'A' || command == 'a') { // get/set file attributes
                if (!read_byte_ex(&val2))
                  op = OP_ERROR;
              }
            }
          }
          else
            op = OP_ERROR;

          if (op != OP_ERROR) {
            if (!read_byte_crc_check())
              op = OP_ERROR;
          }

          if (op != OP_ERROR) {
            if (command == 'L') {  // catalogue list
              if (contains_special_chars((char *) buffer, 0, 1, 1))
                err = RESULT_INVALID_NAME; // invalid name
              else if (length < FAT_MAX_FILE_NAME_LEN) {
                // reset directory pointer
                fat_reset_dir(pDS);
                // prepare file name mask
                prepare_file_name_mask();
              }
            }
            else if (command == 'F') { // create directory
              if (length == 0 || length >= FAT_MAX_FILE_NAME_LEN)
                err = RESULT_PATH_NAME_TOO_LONG; // directory name too long
              else if (contains_special_chars((char *) buffer, 1, 1, 1))
                err = RESULT_INVALID_NAME; // invalid name
              else {
                err = RESULT_OK;

                fat_reset_dir(pDS);
                while (fat_read_dir(pDS, &de)) {
                  if (de.attributes & FAT_ATTRIB_VOLUME)
                    continue;
                  if (strcasecmp(de.long_name, (char *) buffer) == 0) {
                    err = RESULT_FILE_DIR_EXIST; // directory already exist
                    break;
                  }
                }

                if (err == RESULT_OK) {
                  if ((strlen(cur_dir) + 1 + strlen((char *) buffer)) >= FAT_MAX_PATH)
                    err = RESULT_PATH_NAME_TOO_LONG; // too long path
                  else {
                    if (!fat_create_dir(pDS, (char *) buffer, &de))
                      err = RESULT_FATAL_ERROR;
                  }
                }
              }
            }
            else if (command == 'M') { // change directory
              if (length == 0 || length >= FAT_MAX_FILE_NAME_LEN)
                err = RESULT_PATH_NAME_TOO_LONG; // directory name too long
              else if (contains_special_chars((char *) buffer, 1, 1, 1))
                err = RESULT_INVALID_NAME; // invalid name
              else if (strcmp((char *) buffer, ".") == 0) // keep current directory
                err = RESULT_OK;
              else {
                err = RESULT_NOT_FOUND; // not found
                fat_reset_dir(pDS);
                while (fat_read_dir(pDS, &de)) {
                  if (de.attributes & FAT_ATTRIB_VOLUME)
                    continue;
                  else if ((de.attributes & FAT_ATTRIB_DIR) == 0)
                    continue;
                  if (strcasecmp(de.long_name, (char *) buffer) == 0) {
                    if (strcmp((char *) buffer, "..") == 0) { // back one directory level
                      p = strrchr(cur_dir, '/');
                      if (p == NULL)
                        *cur_dir = '\0'; // root
                      else
                        *p = '\0'; // one level back
                      err = RESULT_OK;
                    }
                    else if ((strlen(cur_dir) + 1 + strlen((char *) buffer)) >= FAT_MAX_PATH)
                      err = RESULT_PATH_NAME_TOO_LONG; // too long path
                    else {
                      if (*cur_dir)
                        strcat(cur_dir, "/");
                      strcat(cur_dir, (char *) buffer);
                      err = RESULT_OK;
                    }

                    if (err == RESULT_OK) {
                      pDSnew = fat_open_dir(pFS, &de);
                      if (pDSnew) {
                        fat_close_dir(pDS);
                        pDS = pDSnew;
                        config_write_current_path(cur_dir);
                      }
                      else
                        err = RESULT_FATAL_ERROR;
                    }
                    break;
                  }
                }
              }
            }
            else if (command == 'D') {  // delete file(s) in current directory
              if (length == 0 || length >= FAT_MAX_FILE_NAME_LEN)
                err = RESULT_PATH_NAME_TOO_LONG; // file name too long
              else if (contains_special_chars((char *) buffer, 0, 1, 1))
                err = RESULT_INVALID_NAME; // invalid name
              else {
                val = contains_special_chars((char *) buffer, 1, 0, 0);
                if (val)
                  prepare_file_name_mask();

                err = RESULT_NOT_FOUND; // not found
                fat_reset_dir(pDS);
                while (fat_read_dir(pDS, &de)) {
                  if (de.attributes & (FAT_ATTRIB_VOLUME | FAT_ATTRIB_DIR))
                    continue;
                  if (val) {
                    if (de.attributes & (FAT_ATTRIB_READONLY | FAT_ATTRIB_SYSTEM | FAT_ATTRIB_HIDDEN))
                      continue;
                    // check file name by mask
                    p = strrchr(de.long_name, '.');
                    if (p) {
                      *p = '\0';
                      if (pFile_mask_ext && *pFile_mask_ext && !match_name(pFile_mask_ext, p + 1))
                        continue;
                    }
                    else if (pFile_mask_ext && *pFile_mask_ext != '*')
                      continue;
                    if (*file_mask && !match_name(file_mask, de.long_name))
                      continue;
                    if (p)
                      *p = '.';
                    if (fat_delete_file(pFS, &de))
                      err = RESULT_OK;
                    else {
                      err = RESULT_FILE_DELETE_ERROR;
                      break;
                    }
                  }
                  else {
                    if (strcasecmp(de.long_name, (char *) buffer) == 0) {
                      if ((de.attributes & (FAT_ATTRIB_READONLY | FAT_ATTRIB_SYSTEM | FAT_ATTRIB_HIDDEN)) == 0 && fat_delete_file(pFS, &de))
                        err = RESULT_OK;
                      else
                        err = RESULT_FILE_DELETE_ERROR;
                      break;
                    }
                  }
                }
              }
            }
            else if (command == 'd') {  // delete empty directory
              if (length == 0 || length >= FAT_MAX_FILE_NAME_LEN)
                err = RESULT_PATH_NAME_TOO_LONG; // file name too long
              else if (contains_special_chars((char *) buffer, 1, 1, 1)
                       || strcmp((char *) buffer, "..") == 0
                       || strcmp((char *) buffer, ".") == 0)
                err = RESULT_INVALID_NAME; // invalid name
              else {
                err = RESULT_NOT_FOUND; // not found error code
                fat_reset_dir(pDS);
                while (fat_read_dir(pDS, &de)) {
                  if (de.attributes & (FAT_ATTRIB_VOLUME | FAT_ATTRIB_READONLY | FAT_ATTRIB_SYSTEM | FAT_ATTRIB_HIDDEN))
                    continue;
                  else if ((de.attributes & FAT_ATTRIB_DIR) == 0)
                    continue;
                  if (strcasecmp(de.long_name, (char *) buffer) == 0) {
                    pDSnew = fat_open_dir(pFS, &de);
                    if (!pDSnew)
                      err = RESULT_FATAL_ERROR;
                    else {
                      err = RESULT_OK;
                      pDE = (struct fat_dir_entry_struct *) buffer;
                      val = 1; // can be deleted
                      fat_reset_dir(pDSnew);
                      while (fat_read_dir(pDSnew, pDE)) {
                        if (de.attributes & FAT_ATTRIB_DIR) {
                          if (strcmp(pDE->long_name, "..") == 0)
                            continue;
                          if (strcmp(pDE->long_name, ".") == 0)
                            continue;
                        }
                        val = 0; // dir is not empty
                        break;
                      }
                      fat_close_dir(pDSnew);

                      if (val) {
                        if ((de.attributes & (FAT_ATTRIB_VOLUME | FAT_ATTRIB_READONLY | FAT_ATTRIB_SYSTEM | FAT_ATTRIB_HIDDEN)) > 0 || !fat_delete_dir(pFS, &de))
                          err = RESULT_FILE_DELETE_ERROR;
                      }
                      else
                        err = RESULT_DIR_NOT_EMPTY;
                    }
                    break;
                  }
                }
              }
            }
            else if (command == 'N') {  // rename file/directory
              pb = buffer + 256; // new file/dir name
              if (length == 0 || length >= FAT_MAX_FILE_NAME_LEN
                  || val == 0 || val >= FAT_MAX_FILE_NAME_LEN)
                err = RESULT_PATH_NAME_TOO_LONG; // file name too long
              else if (contains_special_chars((char *) buffer, 1, 1, 1)
                       || strcmp((char *) buffer, "..") == 0
                       || strcmp((char *) buffer, ".") == 0
                       || contains_special_chars((char *) pb, 1, 1, 1)
                       || strcmp((char *) pb, "..") == 0
                       || strcmp((char *) pb, ".") == 0)
                err = RESULT_INVALID_NAME; // invalid name
              else {
                strcpy((char *) (buffer + BUFFER_SIZE - FAT_MAX_FILE_NAME_LEN), (char *) buffer);
                strcpy((char *) buffer, cur_dir);
                strcat((char *) buffer, "/");
                val2 = strlen((char *) buffer);
                if (val2 + length >= FAT_MAX_PATH || val2 + val > FAT_MAX_PATH)
                  err = RESULT_PATH_NAME_TOO_LONG; // path too long
                else {
                  strcat((char *) buffer, (char *) pb);
                  if (fat_get_dir_entry_of_path(pFS, (char *) buffer, &de))
                    err = RESULT_FILE_DIR_EXIST;
                  else {
                    *(buffer + val2) = 0;
                    strcat((char *) buffer, (char *) (buffer + BUFFER_SIZE - FAT_MAX_FILE_NAME_LEN));
                    if (fat_get_dir_entry_of_path(pFS, (char *) buffer, &de)) {
                      if ((de.attributes & FAT_ATTRIB_HIDDEN) == 0 && fat_move_file(pFS, &de, pDS, (char *) pb))
                        err = RESULT_OK;
                      else
                        err = RESULT_FATAL_ERROR;
                    }
                    else
                      err = RESULT_NOT_FOUND;
                  }
                }
              }
            }
            else if (command == 'r' || command == 'o' || command == 'c' || command == 'a') {
              // insert file into ROM Module | open file for reading | create and open file for saving | get/set file attributes
              if (length >= FAT_MAX_PATH)
                err = RESULT_PATH_NAME_TOO_LONG; // path too long
              else if (contains_special_chars((char *) buffer, 1, 0, 1))
                err = RESULT_INVALID_NAME; // invalid name
              else if (command == 'r')  // insert file into ROM Module
                err = prepare_rom_file_into_memory((char *) buffer, addr);
              else if (command == 'o') // open file for reading
                err = open_file_for_read((char *) buffer);
              else if (command == 'c') // create and open file for saving
                err = open_file_for_write((char *) buffer);
              else if (command == 'a') // get/set file attributes
                err = set_file_attrs((char *) buffer, &de, val2);
            }
            else if (command == 'R' || command == 'O' || command == 'C' || command == 'A') {
              // insert file into ROM Module | open file for reading | create and open file for saving | get/set file attributes
              if (length >= FAT_MAX_FILE_NAME_LEN)
                err = RESULT_PATH_NAME_TOO_LONG; // name too long
              else if (contains_special_chars((char *) buffer, 1, 1, 1))
                err = RESULT_INVALID_NAME; // invalid name
              else {
                strcpy((char *) (buffer + BUFFER_SIZE - FAT_MAX_FILE_NAME_LEN), (char *) buffer);
                strcpy((char *) buffer, cur_dir);
                strcat((char *) buffer, "/");
                strcat((char *) buffer, (char *) (buffer + BUFFER_SIZE - FAT_MAX_FILE_NAME_LEN));
                if (strlen((char *) buffer) >= FAT_MAX_PATH)
                  err = RESULT_PATH_NAME_TOO_LONG; // path too long
                else if (command == 'R')  // insert file into ROM Module
                  err = prepare_rom_file_into_memory((char *) buffer, addr);
                else if (command == 'O') // open file for reading
                  err = open_file_for_read((char *) buffer);
                else if (command == 'C')
                  err = open_file_for_write((char *) buffer);
                else if (command == 'a') // get/set file attributes
                  err = set_file_attrs((char *) buffer, &de, val2);
              }
            }
            else if (command == 'B') {
              // write block to file
              err = RESULT_FILE_NOT_OPENED; // file not opened
              if (pWorkWrFile) {
                val16 = (uint16_t) length;
                if (length == 0)
                  val16 = 256;
                vals16 = fat_write_file(pWorkWrFile, buffer, val16);
                if (vals16 < 0)
                  err = RESULT_FATAL_ERROR; // fatal error
                else if (vals16 != val16)
                  err = RESULT_DISK_FULL;
                else
                  err = RESULT_OK;
                if (last_block || err != RESULT_OK) {
                  fat_close_file(pWorkWrFile);
                  pWorkWrFile = NULL;
                }
              }
            }

            if (!(op & OP_SEND_DATA))
              send_byte(err); // send error code
          }
        }

        if (op & OP_SEND_DATA) {  // ** SEND DATA **
          if (command == '@') {     // get FW version
            length = 1;
            *buffer = (uint8_t) FW_VERSION;
          }
          else if (command == 'K') {  // get actual SD card path
            length = strlen(cur_dir);
            *buffer = (uint8_t) length;
            if (length)
              strcpy((char *) (buffer + 1), cur_dir);
            length++;
          }
          else if (command == 'L') {  // catalogue list
            pb = buffer + 1;
            uint32_t size = 0;
            while (fat_read_dir(pDS, &de)) {
              if (de.attributes & FAT_ATTRIB_VOLUME)
                continue;
              else if (de.attributes & FAT_ATTRIB_DIR) {
                if (strcmp(de.long_name, ".") == 0)
                  continue;
              }
              else {
                p = strrchr(de.long_name, '.');
                if (p) {
                  *p = '\0';
                  if (pFile_mask_ext && *pFile_mask_ext && !match_name(pFile_mask_ext, p + 1))
                    continue;
                }
                else if (pFile_mask_ext && *pFile_mask_ext != '*')
                  continue;
                if (*file_mask && !match_name(file_mask, de.long_name))
                  continue;
                if (p)
                  *p = '.';
              }
              strcpy((char *) pb, de.long_name);
              pb += strlen(de.long_name);
              if (!(de.attributes & FAT_ATTRIB_DIR))
                size = de.file_size;
              break;
            }

            *pb++ = de.attributes & 0x37; // ADxSHR
            *pb++ = size & 0xFF;
            *pb++ = (size >> 8) & 0xFF;
            *pb++ = (size >> 16) & 0xFF;
            *pb++ = (size >> 24) & 0xFF;
            length = pb - buffer;
            *buffer = (length - 6);
          }
          else if (command == 'O' || command == 'o') {  // open file for reading
            uint32_t size = pWorkRdFile->dir_entry.file_size;
            pb = buffer;
            *pb++ = size & 0xFF;
            *pb++ = (size >> 8) & 0xFF;
            *pb++ = (size >> 16) & 0xFF;
            *pb++ = (size >> 24) & 0xFF;
            length = 4;
          }
          else if (command == 'P') {  // prepare next 4 kB from opened file
            val16 = 0;
            if (pWorkRdFile) {
              val = is_memory_bank_1(); // remember actual bank
              switch_memory_bank_1();
              err = memory_write_file_content(pWorkRdFile, MEM_FILE_BUFFER_ADDR, MEM_FILE_BUFFER_SIZE, &val16);
              // return original bank
              if (!val)
                switch_memory_bank_0();
            }
            else
              err = RESULT_FILE_NOT_OPENED;
            pb = buffer;
            *pb++ = val16 & 0xFF;
            *pb = (val16 >> 8) & 0xFF;
            length = 2;
          }
          else if (command == 'A' || command == 'a') {  // get/set file attributes
            uint32_t size = de.file_size;
            pb = buffer;
            *pb++ = de.attributes & 0x37; // ADxSHR
            *pb++ = size & 0xFF;
            *pb++ = (size >> 8) & 0xFF;
            *pb++ = (size >> 16) & 0xFF;
            *pb++ = (size >> 24) & 0xFF;
            length = 5;
          }

          send_byte(err); // send error code

          pb = buffer;    // send prepared response from buffer
          crc = 0xAA;
          while (length) {
            val = *pb++;
            crc ^= val;
            send_byte(val);
            length--;
          }
          send_byte(crc); // send CRC
        }
      }
    }

    if (pDS) {
      fat_close_dir(pDS);
      pDS = NULL;
    }
  }
}
//-----------------------------------------------------------------------------
void prepare_file_name_mask()
{
  uint8_t fAst, bt, *pbb;
  char *pc = file_mask;

  // clear file name mask
  *file_mask = '\0';
  pFile_mask_ext = NULL;

  if (!*buffer) {
    // prepare default mask `*.*`
    *pc++ = '*';
    *pc++ = '\0';
    pFile_mask_ext = pc;
    *pc++ = '*';
    *pc = '\0';
    return;
  }
  else {
    pbb = (uint8_t *) strrchr((char *) buffer, '.');  // find extension
    if (buffer == pbb) { // dot as first character
      // prepare `*` as name part
      *pc++ = '*';
    }
  }

  // copy received mask into `file_mask` string and remove duplicate asterisks
  fAst = 0;
  pbb = buffer;
  while (*pbb) {
    bt = *pbb++;              // read char
    if (bt == '*') {          // asterisk?
      if (fAst)               // if previous char was asterisk,
        continue;             // skip repeated asterisk
      fAst = 1;               // flag - was asterisk
    }
    else
      fAst = 0;               // clear asterisk flag
    *pc++ = tolower(bt);      // store char into mask - lower case
  }
  *pc = '\0';                 // mask end

  // check extension
  pFile_mask_ext = strrchr(file_mask, '.');  // find extension
  if (pFile_mask_ext) {
    *pFile_mask_ext = '\0';   // mark end of file name without extension
    pFile_mask_ext++;         // begin of the extension
    if (*pFile_mask_ext == '\0') {
      pc = pFile_mask_ext;
      *pc++ = '*';
      *pc = '\0';
    }
  }
  else {
    pc++;
    pFile_mask_ext = pc;
    *pc++ = '*';
    *pc = '\0';
  }
}
//-----------------------------------------------------------------------------
uint8_t contains_special_chars(const char *pName, uint8_t wildcards, uint8_t slash, uint8_t unallowed)
{
  if (wildcards) {
    p = strchr(pName, '*');
    if (!p)
      p = strchr(pName, '?');
    if (p)
      return 1;
  }

  if (slash) {
    p = strchr(pName, '/');
    if (!p)
      p = strchr(pName, '\\');
    if (p)
      return 1;
  }

  if (unallowed) {
    p = strchr(pName, ':');
    if (!p)
      p = strchr(pName, '"');
    if (!p)
      p = strchr(pName, '<');
    if (!p)
      p = strchr(pName, '>');
    if (!p)
      p = strchr(pName, '|');
    if (p)
      return 1;
  }

  return 0;
}
//-----------------------------------------------------------------------------
/**
 * \brief Compare `pName` against `pMask`.
 * \see   http://xoomer.virgilio.it/acantato/dev/wildcard/wildmatch.html
 *
 * \param pMask pMask string
 * \param pName pName string
 *
 * \return 1 if `pName` match with `pMask`; 0 otherwise
 */
uint8_t match_name(const char *pMask, const char *pName)
{
  const char *pN, *pM;
  int8_t star = 0;

loopStart:
  for (pN = pName, pM = pMask; *pN; pN++, pM++) {
    switch (*pM) {
      case '?' :
        break;

      case '*' :
        star = 1;
        pName = pN;
        pMask = pM;
        if (!*++pMask)
          return 1;
        goto loopStart;

      default :
        if (tolower(*pN) != tolower(*pM)) {
          if (!star)
            return 0;
          pName++;
          goto loopStart;
        }
        break;
    } // endswitch
  } // endfor

  if (*pM == '*')
    ++pM;

  return (!*pM);
}
//-----------------------------------------------------------------------------
uint8_t open_file_for_write(const char *pFilePath)
{
  struct fat_dir_entry_struct de;
  struct fat_dir_struct *pDS;
  char *p;
  uint8_t ret = 0;

  if (pWorkWrFile) {
    fat_close_file(pWorkWrFile);
    pWorkWrFile = NULL;
  }

  if (fat_get_dir_entry_of_path(pFS, pFilePath, &de)) {
    if (de.attributes & (FAT_ATTRIB_VOLUME | FAT_ATTRIB_DIR | FAT_ATTRIB_SYSTEM | FAT_ATTRIB_READONLY))
      return RESULT_FILE_DELETE_ERROR;
    // if file exists, delete it first
    if (!fat_delete_file(pFS, &de))
      return RESULT_FATAL_ERROR;
  }

  p = strrchr(pFilePath, '/');  // find last slash
  if (!p)
    return RESULT_NOT_FOUND;
  if (p == pFilePath) {
    // we are on root
    ret = fat_get_dir_entry_of_path(pFS, "/", &de);
  }
  else {
    *p = '\0'; // terminate path without file name
    ret = fat_get_dir_entry_of_path(pFS, pFilePath, &de);
    *p = '/'; // return slash into path
  }
  if (!ret)
    return RESULT_NOT_FOUND;
  pDS = fat_open_dir(pFS, &de);
  if (!pDS)
    return RESULT_FATAL_ERROR;
  ret = fat_create_file(pDS, p + 1, &de);
  fat_close_dir(pDS);
  if (!ret)
    return RESULT_FATAL_ERROR;

  pWorkWrFile = fat_open_file(pFS, &de);
  if (!pWorkWrFile)
    return RESULT_FATAL_ERROR;
  pWorkWrFile->dir_entry.attributes = FAT_ATTRIB_ARCHIVE;

  return RESULT_OK;
}
//-----------------------------------------------------------------------------
uint8_t open_file_for_read(const char *pFilePath)
{
  struct fat_dir_entry_struct dir_entry;
  uint8_t ret = RESULT_NOT_FOUND;

  if (pWorkRdFile) {
    fat_close_file(pWorkRdFile);
    pWorkRdFile = NULL;
  }

  if (fat_get_dir_entry_of_path(pFS, pFilePath, &dir_entry)) {
    pWorkRdFile = fat_open_file(pFS, &dir_entry);
    if (pWorkRdFile != NULL)
      ret = RESULT_OK;
  }

  return ret;
}
//-----------------------------------------------------------------------------
uint8_t set_file_attrs(const char *pFilePath, struct fat_dir_entry_struct *pDE, uint8_t new_attrs)
{
  uint8_t ret = RESULT_NOT_FOUND;

  if (fat_get_dir_entry_of_path(pFS, pFilePath, pDE)) {
    ret = RESULT_OK;
    if (new_attrs != 255) {
      if (!fat_set_file_attributes(pFS, pDE, new_attrs))
        ret = RESULT_FILE_ACCESS_ERROR;
    }
  }

  return ret;
}
//-----------------------------------------------------------------------------
//*****************************************************************************
// SRAM memory functions
//*****************************************************************************
//-----------------------------------------------------------------------------
/*
 * \ingroup main
 * Function loads ROM file from SD card into memory bank 0.
 *
 * \param[in] pRomFilePath Absolute path to the file.
 * \param[in] addr Offset in the memory where will be file loaded.
 * \returns RESULT_OK, RESULT_FATAL_ERROR, RESULT_NOT_FOUND, RESULT_FILE_ACCESS_ERROR
 *          RESULT_FILE_READ_ERROR
*/
uint8_t prepare_rom_file_into_memory(const char *pRomFilePath, uint16_t addr)
{
  struct fat_dir_entry_struct dir_entry;
  uint8_t ret = RESULT_NOT_FOUND;
  uint8_t bk1 = is_memory_bank_1(); // remeber actual bank

  switch_memory_bank_0();

  // open file from SD card
  if (fat_get_dir_entry_of_path(pFS, pRomFilePath, &dir_entry))
    ret = store_file_into_memory(&dir_entry, addr, ROM_BLOCK_SIZE, NULL);
  if (ret == RESULT_FILE_EOF)
    ret = RESULT_OK;

  // return original bank
  if (bk1)
    switch_memory_bank_1();

  return ret;
}
//----------------------------------------------------------------------------
/*
 * \ingroup main
 * Function loads Manager program from SD card root directory into memory bank 1.
 *
 * \returns RESULT_OK, RESULT_FATAL_ERROR, RESULT_NOT_FOUND, RESULT_FILE_ACCESS_ERROR
 *          RESULT_FILE_READ_ERROR
*/
uint8_t prepare_manager_into_memory()
{
  struct fat_dir_entry_struct dir_entry;
  uint8_t ret = RESULT_NOT_FOUND;

  switch_memory_bank_1();

  // open "manager.rmm" manager file from root directory of the SD card
  strcpy_P((char *) buffer, pstr_mngr_file_name);
  if (fat_get_dir_entry_of_path(pFS, (char *) buffer, &dir_entry))
    ret = store_file_into_memory(&dir_entry, 0, MANAGER_SIZE_MAX, NULL);
  if (ret == RESULT_FILE_EOF)
    ret = RESULT_OK;

  return ret;
}
//----------------------------------------------------------------------------
/**
 * \ingroup main
 * Function fills selected memory bank with file content.
 *
 * File is denoted by directory entry and is read from given offset.
 * Up to `count` bytes from file are wtitten into memory from given address
 * `addr`. But not more than ROM_BLOCK_SIZE bytes. If file is smaller than `count`
 * bytes, rest of the 1 kB boundary memory bank is filled with byte 0xFF.
 * If `count` is zero, it becomes size of the file, but max. `ROM_BLOCK_SIZE - addr`.
 *
 * \param[in] dir_entry The file entry to open.
 * \param[in] addr Offset in the memory where will be file loaded.
 * \param[in] count Count of bytes to read/write.
 * \param[out] pWritten pointer to variable that receive number of bytes wtitten; can be NULL
 * \returns RESULT_OK, RESULT_FATAL_ERROR, RESULT_NOT_FOUND, RESULT_FILE_READ_ERROR, RESULT_FILE_EOF
*/
uint8_t store_file_into_memory(struct fat_dir_entry_struct *dir_entry, uint16_t addr, uint16_t count, uint16_t *pWritten)
{
  struct fat_file_struct *pFD;
  uint8_t ret;

  if (dir_entry == NULL)
    return RESULT_FATAL_ERROR;
  pFD = fat_open_file(pFS, dir_entry);
  if (pFD == NULL)
    return RESULT_NOT_FOUND;

  addr = addr & (ROM_BLOCK_SIZE - 1);
  if (count > ROM_BLOCK_SIZE)
    count = ROM_BLOCK_SIZE;
  else if (count == 0) {
    if (dir_entry->file_size > ROM_BLOCK_SIZE)
      count = ROM_BLOCK_SIZE;
    else
      count = dir_entry->file_size;
  }

  ret = memory_write_file_content(pFD, addr, count, pWritten);

  fat_close_file(pFD);

  return ret;
}
//----------------------------------------------------------------------------
uint8_t memory_write_file_content(struct fat_file_struct *pFD, uint16_t addr, uint16_t count, uint16_t *pWritten)
{
  uint16_t endAddr, wrCnt = 0;
  int16_t cnt;
  uint8_t ret = RESULT_OK, *p;

  if (pFD == NULL)
    return RESULT_FATAL_ERROR;

  endAddr = addr + count;
  if (endAddr > ROM_BLOCK_SIZE)
    endAddr = ROM_BLOCK_SIZE;
  if (endAddr & 0x3FF) // MOD 1 kB
    endAddr = (endAddr + 0x3FF) & ~0x3FF;

  while (addr < endAddr) {
    cnt = fat_read_file(pFD, buffer, BUFFER_SIZE);
    if (cnt < 0) {
      ret = RESULT_FILE_READ_ERROR;  // on read error return 0
      break;
    }
    else if (cnt == 0) {
      if (wrCnt == 0)
        ret = RESULT_FILE_EOF;
      // if file is smaller, fill rest of block with 0xFF
      enable_memory_write();
      while (addr < endAddr) {
        memory_write_byte(addr, 0xFF);
        addr++;
      }
      disable_memory_write();
      break;
    }

    p = buffer;
    enable_memory_write();
    while (cnt > 0 && addr < endAddr) {
      memory_write_byte(addr, *p++);
      addr++;
      cnt--;
      wrCnt++;
    }
    disable_memory_write();
  }

  if (pWritten)
    *pWritten = wrCnt;

  return ret;
}
//----------------------------------------------------------------------------
/*
 *  Write a byte to the memory to the given address.
 *  - addr - <0, 16383>
 *  - bt - byte to write
 */
void memory_write_byte(uint16_t addr, uint8_t bt)
{
  // write address
  memory_write_addr_byte(addr >> 8);        // H byte
  memory_write_addr_byte(addr & 0xFF);      // L byte
  // make ALE pulse
  make_ale_pulse();
  // write data byte
  DATABUS_L_DATA &= ~DATABUS_L_MASK;        // L nibble
  DATABUS_L_DATA |= (bt & DATABUS_L_MASK);
  DATABUS_H_DATA &= ~DATABUS_H_MASK;        // H nibble
  DATABUS_H_DATA |= (bt & DATABUS_H_MASK);
  // make -WEM pulse
  make_wem_pulse();
}
//----------------------------------------------------------------------------
/*
 * Init memory control pins.
 */
void memory_init(void)
{
  MEM_A14_DIR |= MEM_A14_BIT;     // output
  MEM_A14_PORT &= ~MEM_A14_BIT;   // A14=0

  MEM_ALE_DIR |= MEM_ALE_BIT;     // output
  MEM_ALE_PORT &= ~MEM_ALE_BIT;   // ALE=0

  MEM_CSW_DIR |= MEM_CSW_BIT;     // output
  MEM_CSW_PORT &= ~MEM_CSW_BIT;   // -CSW=0

  MEM_WEM_DIR |= MEM_WEM_BIT;     // output
  MEM_WEM_PORT &= ~MEM_WEM_BIT;   // -WEM=0

  disable_memory_write();
}
//----------------------------------------------------------------------------
/*
 * `Disconnect` Memory from bus.
 */
void disable_memory_write(void)
{
  // Data Bus to High Impedance State
  DATABUS_L_DIR &= ~DATABUS_L_MASK;   // Data Bus as Input
  DATABUS_H_DIR &= ~DATABUS_H_MASK;
  DATABUS_L_DATA &= ~DATABUS_L_MASK;  // Z-State
  DATABUS_H_DATA &= ~DATABUS_H_MASK;

  MEM_CSW_PORT |= MEM_CSW_BIT;        // -CSW=1
  MEM_ALE_PORT &= ~MEM_ALE_BIT;       // ALE=0
  MEM_WEM_PORT |= MEM_WEM_BIT;        // -WEM=1
}
//----------------------------------------------------------------------------
/*
 * `Connect` Memory to the bus.
 */
void enable_memory_write(void)
{
  MEM_CSW_PORT &= ~MEM_CSW_BIT;     // -CSW=0
  DATABUS_L_DIR |= DATABUS_L_MASK;  // Data Bus as Output
  DATABUS_H_DIR |= DATABUS_H_MASK;
}
//-----------------------------------------------------------------------------
//*****************************************************************************
// SD card functions
//*****************************************************************************
//-----------------------------------------------------------------------------
/*
 *  Initialize SD card.
 *  Open first partition and prepare filesystem structure.
 *  Function do not return until SD card is successfully initialized.
 *  When SD card is not inserted or cannot be initialized, "SD card select" LED
 *  is blinking.
 */
void prepare_sd_card(void)
{
  while (1) {
    close_sd_card();

    if (sd_raw_available() && sd_raw_init()) {
      // open first partition
      pPart = partition_open(sd_raw_read, sd_raw_read_interval,
                            sd_raw_write, sd_raw_write_interval, 0);
      if (!pPart) {
        /* If the partition did not open, assume the storage device
         * is a "superfloppy", i.e. has no MBR.
         */
        pPart = partition_open(sd_raw_read, sd_raw_read_interval,
                              sd_raw_write, sd_raw_write_interval, -1);
      }
      if (pPart) {
        // open file system
        pFS = fat_open(pPart);
        if (pFS)
          break;
      }
    }

    // blink with "SD card select" LED when card is not present
    // or cannot be initialized
    select_card();
    _delay_ms(250);
    unselect_card();
    _delay_ms(250);
    select_card();
    _delay_ms(250);
    unselect_card();
    _delay_ms(1000);
  }
}
//----------------------------------------------------------------------------
/*
 *  Close SD card and release active partition and filesystem structures.
 */
void close_sd_card(void)
{
  if (pPart && pFS)
    sd_raw_sync();

  // close file system
  if (pFS) {
    fat_close(pFS);
    pFS = NULL;
  }

  // close partition
  if (pPart) {
    partition_close(pPart);
    pPart = NULL;
  }
}
//-----------------------------------------------------------------------------
//*****************************************************************************
// UART functions
//*****************************************************************************
//-----------------------------------------------------------------------------
void uart_init(void)
{
  /* set baud rate */
  UBRR0H = UBRRVAL >> 8;
  UBRR0L = UBRRVAL & 0xFF;
  /* set frame format: 8 bit, no parity, 1 stop bit */
  UCSR0C = (1 << UCSZ01) | (1 << UCSZ00);
  /* enable serial receiver and transmitter */
  UCSR0B = (1 << RXEN0) | (1 << TXEN0);
}
//----------------------------------------------------------------------------
uint8_t read_byte(uint8_t *bt, uint32_t to_ms)
{
  // wait for byte to receive, but only `to_ms` milliseconds
  uint32_t next_ticks = timer_get_ticks() + to_ms;
  while (timer_get_ticks() < next_ticks) {
    if (UCSR0A & (1 << RXC0)) {
      *bt = UDR0;
      return 1;
    }
  }

  return 0;
}
//----------------------------------------------------------------------------
uint8_t read_byte_ex(uint8_t *bt)
{
  if (read_byte(bt, WAIT_BYTE_TIMEOUT)) {
    crc ^= *bt;
    return 1;
  }

  send_byte(NAK);
  return 0;
}
//----------------------------------------------------------------------------
uint8_t read_byte_crc_check(void)
{
  uint8_t b;

  if (read_byte(&b, WAIT_BYTE_TIMEOUT)) {
    crc ^= b;
    if (!crc) {
      send_byte(ACK);
      return 1;
    }
  }

  send_byte(NAK);
  return 0;
}
//-----------------------------------------------------------------------------
void send_byte(uint8_t bt)
{
  // wait until transmit buffer is empty
  while (!(UCSR0A & (1 << UDRE0)))
    ;

  UDR0 = bt;
}
//----------------------------------------------------------------------------
//*****************************************************************************
// EEPROM functions
//*****************************************************************************
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
// structure for configuration that is stored in EEPROM
// structure is used only to calculate addresses of structure items
typedef struct {
  char current_path[FAT_MAX_PATH];    // current path
} EE_CONFIG;
//----------------------------------------------------------------------------
static EE_CONFIG *ee_config = NULL;   // dummy pointer for calculating item addresses
//----------------------------------------------------------------------------
#define EE_ITEM_ADDR(item)            (&ee_config->item)
//----------------------------------------------------------------------------
void config_read_current_path(char *path)
{
  eeprom_read_block(path, EE_ITEM_ADDR(current_path), FAT_MAX_PATH);
}
//----------------------------------------------------------------------------
void config_write_current_path(char *path)
{
  eeprom_write_block(path, EE_ITEM_ADDR(current_path), FAT_MAX_PATH);
}
//----------------------------------------------------------------------------
